local MAJOR = "StarLibPluginResourceTools-1.0" 
local MINOR = 14

local PluginResourceTools = LibStub:NewLibrary(MAJOR, MINOR)
if not PluginResourceTools then return end

if not PluginResourceTools.__index then
	PluginResourceTools.__index = PluginResourceTools
end

local scriptProfile = GetCVar("scriptProfile") == "1"
local oldtotalcpu, oldtotalmem
local lastcputime, lastmemtime = GetTime(), GetTime()

local oldmem = {}
local oldcpu = {}
local meminit
local cpuinit

local new, del
do
	local pool = setmetatable({}, {__mode = "k"})
	function new(...)
		local obj = next(pool)
		if obj then
			pool[obj] = nil
		else
			obj = {}
		end
		for i = 1, select("#", ...) do
			obj[i] = select(i, ...)
		end
		return obj
	end
	function del(tbl)
		pool[tbl] = true
	end
end

-- @name PluginResourceTools:New
-- @usage PluginResourceTools:New(visitor, environment, name, config) -- 
-- @parma environment This will be the environment when loadstring is callled.
-- @return A new PluginResourceTools object
function PluginResourceTools:New(environment)
	
	local obj = {}

	setmetatable(obj, self)
	
	environment.GetCPUUsage = self.GetCPUUsage -- returns cpu, perc, cpudiff, totalcpu, totaldiff
	environment.GetMemUsage = self.GetMemUsage -- returns mem, perc, memdiff, totalmem, totaldiff
	environment.GetLatency = self.GetLatency -- returns current latency
	environment.UpdateCPU = self.UpdateCPU -- Update cpu values
	environment.UpdateMem = self.UpdateMem -- Update memory values
	environment.Update = self.Update -- Call UpdateCPU and UpdateMem
	environment.GetFramerate = self.GetFramerate -- return the current frame rate
	environment.scriptProfile = scriptProfile -- Whether CVar("scriptProfile") is true or not
		
	return obj	
	
end

-- @name PluginResourceTools:Update
-- @usage PluginResourceTools:Update()
-- @return Nothing
function PluginResourceTools:Update()
	PluginResourceTools:UpdateMem()
	PluginResourceTools:UpdateCPU()
end

-- @name UpdateCPU
-- @usage PluginResourceTools.UpdateCPU()
-- @return Nothing
function PluginResourceTools.UpdateCPU()
	cpuinit = true
	
	UpdateAddOnCPUUsage()
	
	local totalcpu = 0.001
	
	for i=1, GetNumAddOns() do
		totalcpu = totalcpu + (GetAddOnCPUUsage(i) )
	end	
	
	for i = 1, GetNumAddOns() do
	
		local addon = GetAddOnInfo(i) 
		
		if not oldcpu[addon] then 
			oldcpu[addon] = new(-1, -1, -1, -1, -1)
		end
		
		del(oldcpu[addon])
						
		local cpu = GetAddOnCPUUsage(addon)
		
		local cpuperc = cpu / totalcpu * 100
		local cpudiff = oldcpu[addon] and oldcpu[addon][1] and (cpu - oldcpu[addon][1]) or 0
		local totaldiff = oldtotalcpu and (totalcpu - oldtotalcpu) or 0

		oldcpu[addon] = new()
		oldcpu[addon][1] = cpu
		oldcpu[addon][2] = cpuperc
		oldcpu[addon][3] = cpudiff
		oldcpu[addon][4] = totalcpu
		oldcpu[addon][5] = totaldiff
	end
	oldtotalcpu = totalcpu	
end

local function getUserCPU(addon)
	if addon == "StarTip" or addon == "LCD4WoW" or usestarlibs then
		local tbl = {"StarLibs-1.0", "StarTip", "LCD4WoW", addon}
		local high = 0
		for i, v in ipairs(tbl) do
			if oldcpu[v] and oldcpu[v][1] > high then
				high = oldcpu[v][1]
				addon = v
			end
		end
	end
	return addon
end

-- @name GetCPUUsage
-- @usage PluginResourceTools.GetCPUUsage(addon)
-- @param addon We'll get the cpu usage for this addon.
-- @param usestarlibs Use this to change the addon name to the highest cpu usage between StarTip, LCD4WoW, StarLibs-1.0, and your addon.
-- @return cpu, percent, cpudiff, totalcpu, totaldiff
function PluginResourceTools.GetCPUUsage(addon, usestarlibs)
	if not scriptProfile then return 0, 0, 0, 0, 0 end
	
	assert(type(addon) == "string", "GetCPUUsage: addon name is not a string")
	
	addon = getUserCPU(addon)

	if not cpuinit then PluginResourceTools.Update() end
	
	if not oldcpu[addon] then
		oldcpu[addon] = new(0, 0, 0, 0, 0)
	end
	
	local cpu = oldcpu[addon][1]
	local perc = oldcpu[addon][2]
	local cpudiff = oldcpu[addon][3]
	local totalcpu = oldcpu[addon][4]
	local totaldiff = oldcpu[addon][5]
	
	local cpuperc
	if totaldiff == 0 then
		totaldiff = 100
		cpudiff = 0
	end
	
	if totaldiff < 0 and cpudiff < 0 then
		local total = abs(totaldiff)
		local cpu = abs(cpudiff)
		cpuperc = cpu / total * 100
		cpuperc = cpuperc * -1
	elseif totaldiff >= 0 and cpudiff >= 0 then
		cpuperc = cpudiff / totaldiff * 100
	else
		cpuperc = 0
	end
	
	return cpu, perc, cpudiff, totalcpu, totaldiff, cpuperc
end

-- @name UpdateMem
-- @usage PluginResourceTools.Update()
-- @return Nothing
function PluginResourceTools.UpdateMem() 
	UpdateAddOnMemoryUsage()
	
	meminit = true
	
	if not oldmem then oldmem = {} end
	
	local totalmem = 0.0001
	
	for i=1, GetNumAddOns() do
		totalmem = totalmem + (GetAddOnMemoryUsage(i) )
	end	

	for i = 1, GetNumAddOns() do	
		local addon = GetAddOnInfo(i) 
		
		if not oldmem[addon] then 
			oldmem[addon] = new(0, 0, 0, 0, 0)
		end
		
		del(oldmem[addon])

		local mem = GetAddOnMemoryUsage(addon)
		local memperc = mem / totalmem * 100
		local memdiff = oldmem[addon][1] and (mem - oldmem[addon][1]) or 0
		local totaldiff = oldtotalmem and (totalmem - oldtotalmem) or 0
				
		oldmem[addon] = new() 
		oldmem[addon][1] = mem
		oldmem[addon][2] = memperc
		oldmem[addon][3] = memdiff
		oldmem[addon][4] = totalmem
		oldmem[addon][5] = totaldiff
	end
	oldtotalmem = totalmem
end

local function getUserMem(addon)
	if addon == "StarTip" or addon == "LCD4WoW" or usestarlibs then
		local tbl = {"StarLibs-1.0", "StarTip", "LCD4WoW", addon}
		local high = 0
		for i, v in ipairs(tbl) do
			if oldmem[v] and oldmem[v][1] > high then
				high = oldmem[v][1]
				addon = v
			end
		end
	end
	return addon
end

-- @name GetMemUsage
-- @usage PluginResourceTools.GetMemUsage(addon)
-- @param addon We'll get the mem usage for this addon
-- @param usestarlibs Use this to change the addon name to the highest mem usage between StarTip, LCD4WoW, StarLibs-1.0, and your addon.
-- @return mem, percent, memdiff, totalmem, totaldiff
function PluginResourceTools.GetMemUsage(addon, userstarlibs)
	assert(type(addon) == "string", "GetCPUUsage: addon name is not a string")	

	if not meminit then PluginResourceTools.UpdateMem() end
	
	addon = getUserMem(addon)

	if not oldmem[addon] then
		oldmem[addon] = new(0, 0, 0, 0, 0)
	end
	
	local mem = oldmem[addon][1]
	local perc = oldmem[addon][2]
	local memdiff = oldmem[addon][3]
	local totalmem = oldmem[addon][4]
	local totaldiff = oldmem[addon][5]

	local memperc
	if totaldiff == 0 then
		totaldiff = 100
		memdiff = 0
	end
	
	if totaldiff < 0 and memdiff < 0 then
		local total = abs(totaldiff)
		local mem = abs(memdiff)
		memperc = mem / total * 100
		memperc = memperc * -1
	elseif totaldiff >= 0 and memdiff >= 0 then
		memperc = memdiff / totaldiff * 100
	else
		memperc = 0
	end
	
	return mem, perc, memdiff, totalmem, totaldiff, memperc
end


local lastlat = 0
function PluginResourceTools.GetLatency()
	local _, _, latency = GetNetStats()
	if type(latency) == "number" then lastlat = latency else latency = lastlat end
	return latency
end

function PluginResourceTools.GetFramerate()
	return GetFramerate() or 0
end
